//io.cpp
//
#include "costanti.h"
#include "tipo.h"
//#define BOCHS
////////////////////////////////////////////////////////////////////////////////
//    COSTANTI                                                                //
////////////////////////////////////////////////////////////////////////////////


const natl PRIO = 1000;
const natl LIV = LIV_SISTEMA;


////////////////////////////////////////////////////////////////////////////////
//                        CHIAMATE DI SISTEMA USATE                           //
////////////////////////////////////////////////////////////////////////////////

extern "C" natl activate_pe(void f(int), int a, natl prio, natl liv, natb type);
extern "C" void terminate_p();
extern "C" void sem_wait(natl sem);
extern "C" void sem_signal(natl sem);
extern "C" natl sem_ini(int val);
extern "C" natl activate_p(void f(int), int a, natl prio, natl liv);
extern "C" void wfi();	// [9.1]
extern "C" void abort_p();
extern "C" void writevid_n(int off, const unsigned char* vett, int quanti);
extern "C" void attrvid_n(int off, int quanti, unsigned char bg, unsigned char fg, bool blink);
extern "C" void log(log_sev sev, const char* buf, int quanti);
extern "C" void* mem_alloc(natl dim, natl balign = 4);
extern "C" void mem_free(void *p);
extern "C" addr graphic_mode();

extern "C" addr trasforma(addr ind_virt);

extern "C" natl pci_find(natl code, natw i);
extern "C" natl pci_read(natw l, natw regn, natl size);
extern "C" void pci_write(natw l, natw regn, natl res, natl size);

//extern "C" natl le32toh(natl le);

////////////////////////////////////////////////////////////////////////////////
//                      FUNZIONI GENERICHE DI SUPPORTO                        //
////////////////////////////////////////////////////////////////////////////////

// ingresso di un byte da una porta di IO
extern "C" void inputb(ioaddr reg, natb &a);

// uscita di un byte su una porta di IO
extern "C" void outputb(natb a, ioaddr reg);

// ingresso di una word da una porta di IO
extern "C" void inputw(ioaddr reg, natw &a);

// uscita di una word su una porta di IO
extern "C" void outputw(natw a, ioaddr reg);

// ingresso di una long da una porta di IO
extern "C" void inputl(ioaddr reg, natl &a);

// uscita di una long su una porta di IO
extern "C" void outputl(natl a, ioaddr reg);

extern "C" void flog(log_sev sev, const char* fmt, ...);
void *memset(void *dest, int c, unsigned int n);
void *memcpy(void *dest, const void *src, unsigned int n);

////////////////////////////////////////////////////////////////////////////////
//                    GESTIONE DELLE INTERFACCE DI RETE                       //
////////////////////////////////////////////////////////////////////////////////

//Numero di schede ethernet presenti
#define N_ETH 1

/*enum ETH_ERR {
	RX_OK = 0,
	TX_OK = 0,
	RX_ERROR = 1,
	TX_ERROR = 1,
	RX_STATUS_ERROR_OR_WRONG_LEN = 2,
	RX_NOT_ENOUGH_SPACE = 3,
	SYSTEM_ERROR = 10
};*/

//Struttura del descrittore d'interfaccia di rete
struct des_eth {
	//Segistri della scheda di rete
	ioaddr CMD;		//8
	ioaddr MPC;		//32
	ioaddr IMR;		//16
	ioaddr ISR;		//16
	
	ioaddr RBSTART;	//32
	ioaddr CAPR;	//16
	ioaddr RCR;		//32
	
	ioaddr TSD0;	//32
	ioaddr TSAD0;	//32
	ioaddr TCR;		//32
	
	//Semafori di sincronizzazione e di mutua esclusione
	int sincr;
	int mutex;
	
	ETH_ERR tx_err; //Variabile per la propagazione dell'errore in trasmissione
	ETH_ERR rx_err; //Variabile per la propagazione dell'errore in ricezione
	
	natb* u_rx_buf; //Buffer (livello utente) ricezione
	natl u_rx_quanti;  //quanti bytes devo ricevere
	
	natb* u_tx_buf; //Buffer (livello utente) invio
	natl u_tx_quanti; //quanti bytes devo trasmettere
	
	//Piedino del controllore su cui arrivano le interruzioni, questo valore
	//verrà aggiornato in fase di inizializzazione della scheda leggendolo dallo
	//spazio di configurazione PCI.
	int irq;
};
extern "C" des_eth deseth[N_ETH];

//Funzione per il reset della scheda di rete
bool eth_reset(int interf) {
	des_eth* p = &deseth[interf];
		
	//Invio del comando di reset
	outputb(0x10, p->CMD);
	natb cmd;
	inputb(p->CMD, cmd);
	for (int i=0; i<1000; i++); //attesa rudimentale
	
	if ((cmd & 0x10) == 0x10) //if (RST == 1) => il reset non è ancora stato completato
		return false;

	return true;
}

//Funzione comune sia alla trasmissione che alla ricezione dei pacchetti:
//disabilita sia ricevitore che trasmettitore
//e maschera tutte le richieste di interruzione che la scheda può generare
void eth_halt(int interf) { 
	des_eth* p = &deseth[interf];
	
	//Azzeramento del Command Register
	outputb(0x00, p->CMD);
	
	//Mascheramento di tutte le interruzioni
	outputw(0x0000, p->IMR);
}

//Durante la fase di probing si effettuano interrogazioni nello spazio di
//configurazione PCI per individuare l'indirizzo base di I/O della scheda
//al quale si sommano gli offset dei registri, e su quale piedino del
//controllore di interruzioneessa essa presenta le richieste di interruzione.
void eth_probe(int interf) {
	des_eth* p = &deseth[interf];
	
	//PCI Class Code Table
	//http://www.acm.uiuc.edu/sigops/roll_your_own/7.c.1.html
	//
	//code[3] = not significant
	//code[2] = PCI Class
	//code[1] = PCI Sub-Class
	//code[0] = prog I/F
	
	natl code = 0x000200ff;
	natl l;
	natw w = 0;
	l = pci_find((natl)code, w);
	
	//Lettura dell'indirizzo di I/O base (dal datasheet)
	ioaddr base = (ioaddr) pci_read(l, 0x10, 2);
	base &= ~0x1;
	
	flog(LOG_DEBUG, "eth%d: PCI-IDE -> I/O base = %x, irq=%d", interf, base, p->irq);
	
	//Inizializzazione del descrittore dell'interfaccia di rete.
	//Non importa che vengano sporcati i campi relativi ad esempio ai semafori
	//od altro, perché la loro inizializzazione verrà fatta in un secondo momento.
	ioaddr* x = (ioaddr*) p;
	
	//La dimensione in word del descrittore di interfaccia è data dalla dimensione
	//totale del descrittore in bytes divisa per la dimensione di una word.
	int wsize = sizeof(des_eth)/(sizeof(natw));
	
	for (int i = 0; i < wsize; i++) {
		*x += base;
		x++;
	}

	//Lettura del piedino del PIC su cui sono generate le richieste d'interruzione
	p->irq = (int) pci_read(l, 0x3c, 1);
}

//Inizializzazione "materiale" della scheda.
void eth_if_init(int interf) {
	des_eth* p = &deseth[interf];

	eth_halt(interf); //serve a stoppare richieste pendenti precedenti

	eth_reset(interf); //reset dell'interfaccia
	
	//Inizializzazione del registro dell'indirizzo (fisico) di trasferimento.
	//
	//In realtà esistono 4 registri diversi di questo tipo, per semplicità
	//ci riferiremo sempre e solo ad uno di questi (TSAD0). In fase di
	//inizializzazione lo settiamo a 0x00000000, quando vorremo trasmettere
	//ci dovremo preoccupare di settarlo con un opportuno indirizzo (fisico)
	//di un pacchetto.
	outputl(0x00000000, p->TSAD0);
	//inizializzazione (a NULL) del puntatore del buffer in trasmissione
	p->u_tx_buf = (natb*)0xffffffff;
	p->u_tx_quanti = 0;
	
	//Inizializzazione del registro per la ricezione.
	//
	//In fase di inizializzazione lo settiamo a 0x00000000 quando vorremo
	//ricevere ci dovremo preoccupare di settarlo con un opportuno indirizzo
	//(fisico) di un'area di memoria adibita alla ricezione.
	outputl(0x00000000, p->RBSTART);
	//inizializzazione (a NULL) del puntatore del buffer in ricezione
	p->u_rx_buf = (natb*)0xffffffff;
	p->u_rx_quanti = 0;
	
	//La scheda dovrebbe inizializzare anche i registri CBR e CAPR, per sicurezza
	//lo facciamo qui con i valori di default suggeriti dai datasheets.
	//
	//Noi non useremo la predisposizione della scheda ad usare un buffer ring,
	//perciò non ci preoccuperemo di utilizzare il registro CBR che sostanzialmente
	//punta, relativamente all'indirizzo contenuto in RBSTART, all'indirizzo di
	//memoria dal quale in poi la scheda può aggiungere pacchetti ricevuti.
	//
	//CAPR: Current Address of Packet Read, contiente l'indirizzo corrente
	//del pacchetto letto.
	//
	//In realtà dato che sono words, sono impropriamente chiamati indirizzi,
	//sarebbe più opportuno chiamarli offsets a partire dall'inidirizzo
	//contenuto nel registro RBSTART.
	outputw(0xfff0, p->CAPR);
		
	//Inizializzazione del registro TCR (Transmit Configuration Register).
	//I parametri che si vanno ad inizializzare sono 2:
	//
	//TCR[25:24] = IFG (InterFrame Gap) -
	//Any value other than 11 will violate the IEEE 802.3 standard.
	//
	//TCR[10:8] = MXDMA -
	//This field sets the maximum size of transmit DMA.
	//Un paper non ufficiale consiglia di impostare la configurazione
	//110 per questo campo che corrisponde a 1024 bytes per ciclo di burst.
	//
	outputl(0x03000600, p->TCR);
	
	//Inizializzazione del registro RCR (Receive Configuration Register)
	//
	//si inizializzano 3 parametri:
	//
	//RCR[15:13] = RXFTH (Rx Fifo THreshold) -
	//noi lo manterremo sempre settato a 111 che indica alla scheda
	//di trasferire nel buffer in memoria, solo quando un pacchetto completo
	//è entrato nel buffer FIFO
	//
	//RCR[12:11] = RBLEN (Rx Buffer LENgth) -
	//This field indicates the size of the Rx ring buffer.
	//Dato che non sfruttiamo i vantaggi offerti dal buffer circolare, e
	//sapendo che un frame ethernet è grande circa 1600 bytes, lo manterremo
	//a 00 che sta a significare
	//8 kbytes + 16 bytes
	//
	//RCR[10:8] = MXDMA -
	//This field sets the maximum size of receive DMA.
	//Noi lo manterremo settato a 111 che indica "il massimo possibile"
	//
	//Occorre inoltr configurare anche i tipi di pacchetti che si vogliono
	//ricevere. Si fa ciò settando dei determinati bits del primo byte di
	//RCR. In particolare mettendo tale byte a 0x0e si accettano pacchetti
	//corretti aventi indirizzi broadcast, multicast e singlecast
	outputl(0x0000e70e, p->RCR);
	
	//Abilitazione alla generazione di interruzioni da parte dell'interfaccia.
	//TER -> 1: Transmit Error, TOK -> 1: Transmit OK
	//RER -> 1: Receive Error, ROK -> 1: Receive OK
	outputw(0x000f, p->IMR);
	
	//Avvio del processo di trasmissione/ricezione.
	//si ottiene azzerando il registro per il conteggio dei pacchetti
	//persi MPC (Missed Packets Counter)
	outputl(0x00000000, p->MPC);
	
	//Abilitazione ricevitore e trasmettitore
	// TE -> 1, RE -> 1;
	outputb(0x0c, p->CMD);
}

void eth_go(des_eth* p) {
	//Abilitazione alla generazione di interruzioni da parte dell'interfaccia.
	//TER -> 1: Transmit Error, TOK -> 1: Transmit OK
	//RER -> 1: Receive Error, ROK -> 1: Receive OK
	//
	natw imr;
	inputw(p->IMR, imr);
	imr |= 0x000f;
	outputw(imr, p->IMR);
	
	//A questo punto l'interfaccia dovrebbe generare un'interruzione non
	//appena il pacchetto sarà traferito.
}

void eth_start_transmit(des_eth* po, natb vetto[], natl quanti) {
	po->u_tx_buf = vetto;
	po->u_tx_quanti = quanti;
	
	//Calcolo indirizzo fisico
	addr ptx_buf = trasforma(po->u_tx_buf);
	
	//Trasferimento indirizzo fisico del pacchetto in TSAD0
	outputl((natl)ptx_buf, po->TSAD0);
	
	natl tsd0;
	inputl(po->TSD0, tsd0);
	tsd0 &= 0xffffd000; // TSD0[13] = OWN -> 0 e TSD0[12:0] = SIZE -> 0
	//Con questa configurazione di bit si abilita il trasferimento del
	//pacchetto dalla memoria fisica al buffer FIFO della scheda
	//
	//Aggiornamento del campo SIZE
	tsd0 |= po->u_tx_quanti;
	//
	//Aggiornamento di TSD0
	outputl(tsd0, po->TSD0);
	
	//Per precauzione in questa fase abilitiamo anche la trasmissione
	natb cmd;
	inputb(po->CMD, cmd);
	cmd |= 0x04;
	outputb(cmd, po->CMD);

	eth_go(po);
}

extern "C" void c_eth_transmit(int eth_if, natb uvetto[], natl uquanti, ETH_ERR& tx_err) {
	des_eth* po;
	po = &deseth[eth_if];

	sem_wait(po->mutex);
	eth_start_transmit(po, uvetto, uquanti);
	sem_wait(po->sincr);
	//qui sappiamo che il processo esterno ha aggiornato il campo
	//tx_err del descrittore di interfaccia per cui possiamo
	//propagarlo a livello utente.
	tx_err = po->tx_err;
	sem_signal(po->mutex);
}

void eth_start_receive(des_eth* pi, natb vetti[], natl quanti) {
	pi->u_rx_buf = vetti;
	pi->u_rx_quanti = quanti;
	
	//calcolo indirizzo fisico di ricezione
	addr prx_buf = trasforma(pi->u_rx_buf);
	
	//inizializzazione del registro appropriato nella scheda
	outputl((natl)prx_buf, pi->RBSTART);
	outputl(0xfff0, pi->CAPR);
	
	//Per precauzione in questa fase abilitiamo anche la ricezione
	natb cmd;
	inputb(pi->CMD, cmd);
	cmd |= 0x08;
	outputb(cmd, pi->CMD);
	
	eth_go(pi);
}

extern "C" void c_eth_receive(int eth_if, natb uvetti[], natl uquanti, ETH_ERR& rx_err) {
	des_eth* pi;
	pi = &deseth[eth_if];

	sem_wait(pi->mutex);
	eth_start_receive(pi, uvetti, uquanti);
	sem_wait(pi->sincr);
	//qui sappiamo che il processo esterno ha aggiornato il campo
	//rx_err del descrittore di interfaccia per cui possiamo
	//propagarlo a livello utente.
	rx_err = pi->rx_err;
	sem_signal(pi->mutex);
}

void ext_eth(int h) {
	des_eth* p = &deseth[h];
	natw isr;
	
	for (;;) {
		inputw(p->ISR, isr);
		
		if (isr == 0xffff)
			goto safe_break;
		
		if (isr != 0)
			outputw(isr, p->ISR);
		
		if ((isr & 0xc07f) == 0)
			goto safe_break;
		
		if (isr & 0x0001) {
			//se (ROK == 1) vuol dire che un pacchetto è stato ricevuto correttamente.
			p->rx_err = RX_OK; // 0
			
			//andiamo a leggere il Command Register per capire se il buffer
			//è pieno o vuoto.
			natb cmd;
			inputb(p->CMD, cmd);
			while ((cmd & 0x01) == 0) {
				//mentre (BUFE == 0), cioè mentre il buffer non è vuoto,
				//cerchiamo di prendere i pacchetti.
				
				//Offset di estrazione dalla coda
				natw capr;
				inputw(p->CAPR, capr);
				capr += 16;
				
				//Ogni pacchetto è preceduto da un header composto da una
				//sezione detta packet status e da un'altra detta packet length
				//Lo header è lungo una parola lunga.
				natl rx_header = *(natl*)p->u_rx_buf;
				//flog(LOG_ERR, "capr = %x, rx_header = %x", capr, rx_header);
				
				
				//adesso è possibile leggere la lunghezza del pacchetto
				//shiftando a destra di 16 rx_status
				natw packet_len = rx_header >> 16;
				
				//l'utente ha allocato poco spazio. dev'essere allocato più spazio.
				if (p->u_rx_quanti < (packet_len + 8)) {
					p->rx_err = RX_NOT_ENOUGH_SPACE; // 3
					goto safe_break;
				}

				if (!(rx_header & 0x00000001) || //errore nella ricezione (ROK == 0)
					packet_len < 64 || //pacchetto minore della lunghezza minimina di un frame ethernet
					packet_len > (1518 + 4) //pacchetto maggiore della lunghezza del frame ethernet + 4 per la vlan encap.
					) { //vuol dire che c'è un errore
					p->rx_err = RX_STATUS_ERROR_OR_WRONG_LEN; // 2
					goto safe_break; //esci dal ciclo con errore
				}
				//non è stato rilevato nessun errore
				//si passa alla ricezione del pacchetto
				
				capr += 4;
				//per finire capr viene allineato alla parola lunga
				capr = (capr + 3) & ~3;
				//e quindi si aggiorna il suo valore anche sulla scheda
				outputw(capr - 16, p->CAPR);
				inputb(p->CMD, cmd);
			}

			//riportiamo a 0 ROK per far capire alla scheda che abbiamo
			isr &= 0xfffe; //finito di gestire l'interruzione
			outputw(isr, p->ISR);
		}
		
		if (isr & 0x0002) { //(RER == 1)
			//se (RER == 1) vuol dire che abbiamo avuto degli errori nella ricezione
			p->rx_err = RX_ERROR; // 1
			
			//riportiamo a 0 RER per far capire alla scheda che abbiamo
			isr &= 0xfffd; //finito di gestire l'interruzione
			outputw(isr, p->ISR);
		}
		
		if (isr & 0x0004) { //(TOK == 1)
			//se (TOK == 1) vuol dire che il trasferimento è stato completato con successo
			//per cui si liberano le strutture per l'immagazzinamento del pacchetto
			p->tx_err = TX_OK; // 0

			//riportiamo a 0 TOK per far capire alla scheda che abbiamo
			isr &= 0xfffc; //finito di gestire l'interruzione
			outputw(isr, p->ISR);
		}
		
		if (isr & 0x0008) { //(TER == 1)	
			//se (TER == 1) vuol dire che abbiamo avuto degli errori nel
			//trasferimento. In maniera semplificata si ipotizza che il motivo
			//del mancato trasferimento, sia dovuto ad un'errata soglia di
			//trasferimento.
			p->tx_err = TX_ERROR; // 1
			
			//preleviamo la vecchia soglia TSD0[21:16] = ERTXTH
			natl tsd0;
			inputl(p->TSD0, tsd0);
			natl ertxth = (tsd0 & 0x003f0000) >> 11; //preleviamo il valore di
			//ertxth shiftando a destra di 16 tsd0 & 0x003f0000, ogni unità di
			//tale campo è considerata moltiplicato per 32 bytes per cui bisognerebbe
			//shiftare a sinistra di 5 per ottenere il corretto valore rappresentato.
			// 16 - 5 = 11.
			//
			//Adesso si può leggere il vecchio valore della soglia in bytes.
			//Tale campo ha un valore massimo raggiungibile che è pari a
			//2048 bytes, perciò nell'andare ad aumentarlo dobbiamo fare attenzione
			//a non eccedere tale "upper-bound".
			if (ertxth < 2016)
				ertxth += 32;
			tsd0 = (tsd0 & 0xffc0ffff) | (ertxth << 11);
			outputl(tsd0, p->TSD0);
				
			//riportiamo a 0 TER per far capire alla scheda che abbiamo
			isr &= 0xfff7; //finito di gestire l'interruzione
			outputw(isr, p->ISR);
		}
		
		if (isr & 0x8000) { //(SERR == 1)
			//(isr == 0x8000) vuol dire che la scheda va resettata
			p->tx_err = SYSTEM_ERROR; // 10
			p->rx_err = SYSTEM_ERROR; // 10
			eth_reset(h);
			eth_if_init(h);
		}

		if (isr == 0)
safe_break:
			//se si esce con il "safe break" si stoppa la scheda.
			eth_halt(h);

		sem_signal(p->sincr);
		wfi();
	}
	
}

bool eth_init() {
	for (int i=0; i < N_ETH; i++) {
		des_eth* p = &deseth[i];
		
		//Fase Probe.
		//
		//In questa fase principalmente vengono "scoperti":
		//
		//il numero del piedino del controllore di interruzione su cui
		//arrivano le richieste di interruzione da parte dell'interfaccia
		//
		//l'indirizzo base di I/O a partire dal quale iniziano i registri
		//della nostra interfaccia di rete.
		eth_probe(i);

		//Adesso inizializziamo i semafori. Lo facciamo dopo la Fase Probe,
		//perché essa va a "sporcare" il descrittore di interfaccia e quindi
		//anche i valori dei nostri semafori.
		
		//Inizializzazione semaforo di mutua esclusione
		if ((p->mutex = sem_ini(1)) == 0) {
			flog(LOG_ERR, "eth%d: impossibile creare mutex", i);
			return false;
		}
		flog(LOG_DEBUG, "eth%d: creato semaforo mutex", i);
		
		//Inizializzazione semaforo di sincronizzazione
		if ((p->sincr = sem_ini(0)) == 0) {
			flog(LOG_ERR, "eth%d: impossibile creare sincr", i);
			return false;
		}
		flog(LOG_DEBUG, "eth%d: creato semaforo sincr", i);
		
		//Fase Attach.
		//
		//Sostanzialmente consiste nel reset dell'interfaccia.
		bool eth_rst;
		
		//il reset va sempre protetto in mutua esclusione!
		sem_wait(p->mutex);
		eth_rst = eth_reset(i);
		sem_signal(p->mutex);
		
		if (!eth_rst) {
			flog(LOG_ERR, "eth%d: impossibile resettare l'interfaccia di rete!", i);
			return false;
		}
		flog(LOG_INFO, "eth%d: reset dell'interfaccia di rete completato.", i);

		//Fase Init
		sem_wait(p->mutex);
		eth_if_init(i);
		sem_signal(p->mutex);
		
		//Inizializziamo i campi di propagazione a livello superiore dell'errore
		p->rx_err = RX_OK; // 0
		p->tx_err = TX_OK; // 0
		
		//Attivazione del processo esterno
		//Ricordiamoci che p->irq è stato inizializzato durante la Fase Probe
		short id = activate_pe(ext_eth, 0, PRIO, LIV_SISTEMA, p->irq);
		if (id == 0) {
			flog(LOG_ERR, "eth%d: impossibile creare proc. esterno", i);
			return false;
		}
		flog(LOG_DEBUG, "eth%d: proc. esterno creato correttamente", i);
	}
	flog(LOG_INFO, "eth: inizializzazione interfacce ethernet terminata con successo.");
	return true;
}


////////////////////////////////////////////////////////////////////////////////
//                    GESTIONE DELLE INTERFACCE SERIALI [9.2]                 //
////////////////////////////////////////////////////////////////////////////////

enum funz { input_n, input_ln, output_n, output_0 };  // [9.2]

struct interfse_reg {	// [9.2]
	ioaddr iRBR, iTHR, iLSR, iIER, iIIR;
};

struct des_se {		// [9.2]
	interfse_reg indreg;
	natl mutex;
	natl sincr;
	natl cont;
	addr punt;
	funz funzione;
	natb stato;
};

const natl S = 2;
extern "C" des_se com[S];	// [9.2]

void input_com(des_se* p_des);	// [9.2]
void output_com(des_se* p_des);	// [9.2]
void estern_com(int i) // [9.2]
{
	natb r;
	des_se *p_des;
	p_des = &com[i];
	for(;;) {
		inputb(p_des->indreg.iIIR, r);
		if ((r&0x06) == 0x04) 
			input_com(p_des);
		else if ((r&0x06) == 0x02)
			output_com(p_des);
		wfi();
	}
}

void startse_in(des_se *p_des, natb vetti[], natl quanti, funz op); // [9.2.1]
extern "C" void c_readse_n(natl serial, natb vetti[], natl quanti, natb& errore) // [9.2.1]
{
	des_se *p_des;

	// (* le primitive non devono mai fidarsi dei parametri
	if (serial >= S) {
		flog(LOG_WARN, "readse_n con serial=%d", serial);
		abort_p();
	}
	// *)

	p_des = &com[serial];
	sem_wait(p_des->mutex);
	startse_in(p_des, vetti, quanti, input_n);
	sem_wait(p_des->sincr);
	errore = p_des->stato;
	sem_signal(p_des->mutex);
}

extern "C" void c_readse_ln(natl serial, natb vetti[], int& quanti, natb& errore)
{
	des_se *p_des;

	// (* le primitive non devono mai fidarsi dei parametri
	if (serial >= S) {
		flog(LOG_WARN, "readse_ln con serial=%d", serial);
		abort_p();
	}
	// *)

	p_des = &com[serial];
	sem_wait(p_des->mutex);
	startse_in(p_des, vetti, 80, input_ln);
	sem_wait(p_des->sincr);
	quanti = p_des->cont;
	errore = p_des->stato;
	sem_signal(p_des->mutex);
}

extern "C" void go_inputse(ioaddr i_ctr);
void startse_in(des_se *p_des, natb vetti[], natl quanti, funz op) // [9.2.1]
{
	p_des->cont = quanti;
	p_des->punt = vetti;
	p_des->funzione = op;
	go_inputse(p_des->indreg.iIER);
}

extern "C" void halt_inputse(ioaddr i_ctr);
void input_com(des_se *p_des) // [9.2.1]
{
	natb c; bool fine;
	fine = false;

	halt_inputse(p_des->indreg.iIER);

	inputb(p_des->indreg.iLSR, c);

	p_des->stato = c & 0x1e;
	if (p_des->stato != 0)
		fine = true;
	else {
		inputb(p_des->indreg.iRBR, c);
		if (p_des->funzione == input_n) {
			*static_cast<natb*>(p_des->punt) = c; // memorizzazione
			p_des->punt = static_cast<natb*>(p_des->punt) + 1;
			p_des->cont--;
			if(p_des->cont == 0)
				fine = true;
		} else {
			if (p_des->funzione == input_ln) {
				if(c == '\r' || c == '\n') {
					fine = true;
					p_des->cont = 80 - p_des->cont;
				} else {
					*static_cast<natb*>(p_des->punt) = c; // memorizzazione
					p_des->punt = static_cast<natb*>(p_des->punt) + 1;
					p_des->cont--;
					if (p_des->cont == 0) {
						fine = true;
						p_des->cont = 80;
					}
				}
			}
		}
	}

	if(fine == true) {
		*static_cast<natb*>(p_des->punt) = 0;	// carattere nullo
		sem_signal(p_des->sincr);
	} else
		go_inputse(p_des->indreg.iIER);
}

void startse_out(des_se *p_des, natb vetto[], natl quanti, funz op);
extern "C" void c_writese_n(natl serial, natb vetto[], natl quanti)	// [9.2.2]
{
	des_se *p_des;

	// (* le primitive non devono mai fidarsi dei parametri
	if (serial >= S) {
		flog(LOG_WARN, "writese_n con serial=%d", serial);
		abort_p();
	}
	// *)

	p_des = &com[serial];
	sem_wait(p_des->mutex);
	startse_out(p_des, vetto, quanti, output_n);
	sem_wait(p_des->sincr);
	sem_signal(p_des->mutex);
}

extern "C" void c_writese_0(natl serial, natb vetto[], natl &quanti)
{
	des_se *p_des;

	// (* le primitive non devono mai fidarsi dei parametri
	if (serial >= S) {
		flog(LOG_WARN, "writese_0 con serial=%d", serial);
		abort_p();
	}
	// *)

	p_des = &com[serial];
	sem_wait(p_des->mutex);
	startse_out(p_des, vetto, 0, output_0);
	sem_wait(p_des->sincr);
	quanti = p_des->cont;
	sem_signal(p_des->mutex);
}

extern "C" void go_outputse(ioaddr i_ctr);
void startse_out(des_se *p_des, natb vetto[], natl quanti, funz op) // [9.2.2]
{
	p_des->cont = quanti;
	p_des->punt = vetto;
	p_des->funzione = op;
	go_outputse(p_des->indreg.iIER);
	output_com(p_des); 
}

extern "C" void halt_outputse(ioaddr i_ctr);
void output_com(des_se *p_des)	// [9.2.2]
{
	natb c; bool fine;
        fine = false;

	if (p_des->funzione == output_n) {
		p_des->cont--;
		if(p_des->cont == 0) {
			fine = true;
			halt_outputse(p_des->indreg.iIER);
		}
		c = *static_cast<natb*>(p_des->punt); //prelievo
		outputb(c, p_des->indreg.iTHR);
		p_des->punt = static_cast<natb*>(p_des->punt) + 1; 
	} else if (p_des->funzione == output_0) {
		c = *static_cast<natb*>(p_des->punt); //prelievo
		if (c == 0) {
			fine = true;
			halt_outputse(p_des->indreg.iIER);
		} else {
			outputb(c, p_des->indreg.iTHR);
			p_des->cont++;
			p_des->punt = static_cast<natb*>(p_des->punt) + 1; 
		}
	}

	if (fine == true)
		sem_signal(p_des->sincr);

}

// ( inizializzazione delle interfacce seriali
extern "C" void com_setup(void);	// vedi "io.S"
// interruzioni hardware delle interfacce seriali
int com_irq[S] = { 4, 3 };

bool com_init()
{
	des_se *p_des;
	natw id;
	natl i, com_base_prio = PRIO;

	com_setup();

	for(i = 0; i < S; ++i) {
		p_des = &com[i];

		if ( (p_des->mutex = sem_ini(1)) == 0xFFFFFFFF) {
			flog(LOG_ERR, "com: impossibile creare mutex");
			return false;
		}
		if ( (p_des->sincr = sem_ini(0)) == 0xFFFFFFFF) {
			flog(LOG_ERR, "com: impossibile creare sincr");
			return false;
		}

		id = activate_pe(estern_com, i, com_base_prio - i, LIV, com_irq[i]);
		if (id == 0xFFFFFFFF) {
			flog(LOG_ERR, "com: impossibile creare proc. esterno");
			return false;
		}

	}
	flog(LOG_INFO, "com: inizializzate %d seriali", S);
	return true;
}
// )


////////////////////////////////////////////////////////////////////////////////
//                         GESTIONE DELLA TASTIERA                            //
////////////////////////////////////////////////////////////////////////////////


struct des_kbd {
	bool escape;
	int pause;
	ioaddr iRBR;
	natl id;
} kbd;

natb pausecode[6] = { 0xe1, 0x1d, 0x45, 0xe1, 0x9d, 0xc5 };

natw kbd_read()
{
	natb c;

	inputb(kbd.iRBR, c);

	if (c == pausecode[kbd.pause]) {
		if (++kbd.pause >= 6) {
			kbd.pause = 0;
			return 0xe17f;
		}
		return 0;
	} else {
		kbd.pause = 0;
	}

	if (c == 0xe0) {
		kbd.escape = true;
		return 0;
	}
	
	if (kbd.escape) {
		kbd.escape = false;
		return 0xe000 | c;
	}

	return c;
}

// smistatore dei caratteri dalla tastiera reale alle tastiere virtuali
void vkbd_putchar(natw code);
void estern_kbd(int h)
{
	natw code;

	for(;;) {
		code = kbd_read();
		if (code) 
			vkbd_putchar(code);				
		wfi();
	}
}
// Interruzione hardware della tastiera
const int KBD_IRQ = 1;

bool kbd_init()
{

	kbd.escape = false;
	kbd.pause = 0;
	kbd.iRBR = (ioaddr)0x60;

	if ( (kbd.id = activate_pe(estern_kbd, 0, PRIO, LIV, KBD_IRQ)) == 0xFFFFFFFF) {
		flog(LOG_ERR, "kbd: impossibile creare estern_kbd");
		return false;
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////
//    TASTIERE VIRTUALI                                                      //
///////////////////////////////////////////////////////////////////////////////

// sottrazione modulo 'mod' (il '%' del C/C++ calcola il valore sbagliato)
// p1 e p2 devono essere compresi tra 0 e mod-1
inline int circular_sub(int p1, int p2, int mod)
{
	int dist = p1 - p2;
	if (dist < 0) dist += mod;
	return dist;
}

// somma modulo 'mod'
inline int circular_sum(int p1, int p2, int mod)
{
	return (p1 + p2) % mod;
}



const int VKBD_BUF_SIZE = 20;
const int MAX_VKBD = 5;
const int VKBD_CONSOLE = MAX_VKBD - 1;

struct des_vkbd {
	natl mutex;
	natl intr;
	bool intr_enabled;
	bool intr_pending;
	int intr_waiting;
	natw buf[VKBD_BUF_SIZE];
	int first;
	int last;
	int nchar;
	natb leds;
};

des_vkbd array_desvkbd[MAX_VKBD];
des_vkbd *vkbd_active;

bool vkbd_init()
{
	for (int i = 0; i < MAX_VKBD; i++) {
		des_vkbd *k = &array_desvkbd[i];

		if ( (k->mutex = sem_ini(1)) == 0xFFFFFFFF) {
			flog(LOG_ERR, "vkbd: impossibile creare mutex");
			return false;
		}
		if ( (k->intr = sem_ini(0)) == 0xFFFFFFFF) {
			flog(LOG_ERR, "vkbd: impossibile creare intr");
			return false;
		}
		k->intr_enabled = false;
		k->intr_pending = false;
		k->intr_waiting = 0;
		k->first = k->last = k->nchar = 0;
	}
	flog(LOG_INFO, "vkbd: inizializzate %d tastiere virtuali", MAX_VKBD);
	return true;
}

extern "C" void c_vkbd_wfi(int v)
{
	if (v < 0 || v >= MAX_VKBD) {
		flog(LOG_WARN, "(wfi) vkbd inesistente: %d", v);
		abort_p();
	}

	des_vkbd *k = &array_desvkbd[v];

	sem_wait(k->mutex);
	while (!k->intr_pending) {
		k->intr_waiting++;
		sem_signal(k->mutex);
		sem_wait(k->intr);
		sem_wait(k->mutex);
	}
	sem_signal(k->mutex);
}


extern "C" void c_vkbd_switch(int v);
extern "C" void c_vmon_switch(int v);
void vkbd_nuovo_car(des_vkbd* k, natw code, bool clear = false)
{
	sem_wait(k->mutex);

	if (!clear && k->nchar >= VKBD_BUF_SIZE)
		goto out;

	if (clear) 
		k->first = k->last = k->nchar = 0;

	k->buf[k->last] = code;
	k->last = circular_sum(k->last, 1, VKBD_BUF_SIZE);
	k->nchar++;

	if (k->intr_enabled && !k->intr_pending) {
		k->intr_pending = true;
		if (k->intr_waiting > 0) {
			k->intr_waiting--;
			sem_signal(k->intr);
		}
	}
out:
	sem_signal(k->mutex);
}

void vkbd_putchar(natw code)
{
	if (vkbd_active)
		vkbd_nuovo_car(vkbd_active, code);
}

extern "C" void c_vkbd_send(int v, natw code, bool clear)
{
	if (v < -2 || v >= MAX_VKBD) {
		flog(LOG_WARN, "(send) vkbd inesistente: %d", v);
		abort_p();
	}

	des_vkbd *k;

	if (v == -1) {
		for (int i = 0; i < MAX_VKBD; i++) {
			k = &array_desvkbd[i];
			vkbd_nuovo_car(k, code, clear);
		}
	} else {
		if (v == -2) v = VKBD_CONSOLE;
		k = &array_desvkbd[v];
		vkbd_nuovo_car(k, code, clear);
	}
}

extern "C" natw c_vkbd_read(int v)
{
	natw code = 0;

	if (v < 0 || v >= MAX_VKBD) {
		flog(LOG_WARN, "(read) vkbd inesistente: %d", v);
		abort_p();
	}

	des_vkbd *k = &array_desvkbd[v];

	sem_wait(k->mutex);

	if (k->nchar <= 0)
		goto out;

	code = k->buf[k->first];
	k->first = circular_sum(k->first, 1, VKBD_BUF_SIZE);
	k->nchar--;

	k->intr_pending = false;

	if (k->nchar > 0 && k->intr_enabled) {
		k->intr_pending = true;
		if (k->intr_waiting > 0) {
			k->intr_waiting--;
			sem_signal(k->intr);
		}
	}
out:
	sem_signal(k->mutex);
	return code;
}

extern "C" void c_vkbd_intr_enable(int v, bool enable)
{
	if (v < 0 || v >= MAX_VKBD) {
		flog(LOG_WARN, "(intr_en) vkbd inesistente: %d", v);
		abort_p();
	}

	des_vkbd *k = &array_desvkbd[v];

	sem_wait(k->mutex);
	if (enable) {
		if (!k->intr_enabled && k->nchar > 0) {
			k->intr_pending = true;
			if (k->intr_waiting > 0) {
				k->intr_waiting--;
				sem_signal(k->intr);
			}
		}
		k->intr_enabled = true;
	} else {
		k->intr_enabled = false;
	}

	sem_signal(k->mutex);
}

const natl VKBD_LED_CAPSLOCK  = 1L;
const natl VKBD_LED_NUMLOCK   = 2L;
const natl VKBD_LED_SCROLLOCK = 4L;

extern "C" void kbd_set_leds(natb leds);

extern "C" void c_vkbd_leds(int v, natb led, bool on)
{
	if (v < 0 || v >= MAX_VKBD) {
		flog(LOG_WARN, "(leds) vkbd inesistente: %d", v);
		abort_p();
	}

	des_vkbd *k = &array_desvkbd[v];
	if (on) 
		k->leds |= led;
	else
		k->leds &= ~led; 

	if (k == vkbd_active)
		kbd_set_leds(k->leds);
}


extern "C" void c_vkbd_switch(int v)
{
	if (v < 0 || v >= MAX_VKBD) {
		flog(LOG_WARN, "(switch) vkbd inesistente: %d", v);
		abort_p();
	}

	des_vkbd *k = &array_desvkbd[v];

	vkbd_active = k;
	kbd_set_leds(vkbd_active->leds);
}


///////////////////////////////////////////////////////////////////////////////
//  MONITOR VIRTUALI                                                         //
///////////////////////////////////////////////////////////////////////////////

const natl SQRT_NUM_SCREEN = 2;
const natl NUM_SCREEN = 4;

const natl MAX_VMON = NUM_SCREEN;
natw *VIDEO_MEM_BASE = reinterpret_cast<natw *>(0x000b8000);
const int VIDEO_MEM_SIZE = 80 * 25;

const natl FONT_H = 16;
const natl FONT_W = 8;
const natb ch_bitmap[][FONT_H] = {
#include "font.txt"
};
const natl FONT_NUM = sizeof(ch_bitmap) / FONT_H;
natw max_off;

// descrittore di monitor virtuale
struct des_vmon {
	natw* buf;
	natl ch_x, ch_y;
	// text mode
	natw* video;
	int cursor;
	int cursor_size;
	// graph mode
	natl x, y;
	natl w, h;
	natl ubar, dbar;
};

des_vmon array_desvmon[MAX_VMON];
des_vmon *vmon_active;

extern "C" void console_cursor(unsigned int);
extern "C" void console_set_cursor_size(int size);

natb* framebuffer = 0;


static inline void write_pixel(natl x, natl y, natb c) {
	framebuffer[y * VID_W + x] = c;
}

static inline void write_screen(des_vmon* t, natl x, natl y, natb c) {
	write_pixel(t->x + x, t->y + y, c);
}

void write_gc(natl x, natl y, natw v) {
	natb fg = (v >> 8) & 0x0F;
	natb bg = (v >> 12) & 0x07;
	natb c = v;
	if (c >= FONT_NUM) 
		return;
	for (natl i = 0; i < FONT_H; i++) {
		natb w = ch_bitmap[c][i];
		for (natl j = 0; j < FONT_W; j++) {
			write_pixel(x + j, y + i, (w & 0x80) ? fg: bg);
			w <<= 1;
		}
	}
}

void color_bars(des_vmon *t, natb color) {
	//upbar
	for (natl k = 0; k < t->ubar; k++)
		for (natl j = 0; j < t->w; j++) 
			write_screen(t, j, k, color);

	//downbar
	natl downbar = t->h - t->dbar;
	for (natl k = downbar; k < t->h; k++)
		for (natl j = 0; j < t->w; j++) 
			write_screen(t, j, k, color);
}

void framebuffer_cursor(des_vmon *t) {
	natl riga = t->cursor / t->ch_x;
	if (riga > t->ch_y)
		return;
	natl colonna = t->cursor % t->ch_x;
	for (natl i = FONT_H - 1; i >= t->cursor_size; i--) 
		for (natl j = 0; j < FONT_W; j++) 
			framebuffer[(t->y + t->ubar + riga * FONT_H + i) * VID_W +
				    t->x + colonna * FONT_W + j] ^= 0x0F;

}

void framebuffer_update_cursor(des_vmon *t, natl off) {
	if (off == t->cursor)
		return;
	framebuffer_cursor(t);
	t->cursor = off;
	framebuffer_cursor(t);
}


bool vmon_init()
{
	framebuffer = static_cast<natb*>(graphic_mode());
	if (framebuffer)
		flog(LOG_INFO, "modalita' grafica");

	for (natl s = 0; s < MAX_VMON; s++) {
		des_vmon *t = &array_desvmon[s];
		t->cursor = 0;
		t->cursor_size = 14;
		
		if (framebuffer) {
			t->w = VID_W / SQRT_NUM_SCREEN;
			t->h = VID_H / SQRT_NUM_SCREEN;
	 		t->x = t->w * (s % SQRT_NUM_SCREEN);
			t->y = t->h * (s / SQRT_NUM_SCREEN);
			t->ubar = 32;
			t->dbar = 16;
			t->ch_x = t->w / FONT_W;
			t->ch_y = (t->h - t->ubar - t->dbar) / FONT_H;

			color_bars(t, 0x09);	
			////divisori verticali
			//for (int k = 0; k < t->h; k++) {
			//	write_screen(s, 0, k, 0x0B);
			//	write_screen(s, t->w - 1, k, 0x0B);
			//}

			////divisori orizzontali
			//for (int k = 0; k < t->w; k++) {
			//	write_screen(s, k, t->h - 1, 0x0B);
			//	write_screen(s, k, 0, 0x0B);
			//}
			framebuffer_cursor(t);
		} else {
			t->ch_x = 80;
			t->ch_y = 25;
		}
		t->video = t->buf = static_cast<natw*>(mem_alloc((t->ch_x * t->ch_y) * sizeof(natw)));
	} 
	flog(LOG_INFO, "vmon: inizializzati %d monitor virtuali", MAX_VMON);
	return true;
}

static void vmon_put(int v, int off, natw code)
{
	des_vmon *t = &array_desvmon[v];
	if (framebuffer) {
		int riga = off / t->ch_x;
		int colonna = off % t->ch_x;
		write_gc(t->x + colonna * FONT_W,
		         t->y + riga * FONT_H + t->ubar, code);
		if (off == t->cursor)
			framebuffer_cursor(t);
	}
	t->video[off] = code;
}

static inline natw vmon_get(int v, int off)
{
	return array_desvmon[v].video[off];
}

extern "C" void c_vmon_write_n(int v, int off, natw vetti[], int quanti)
{
	if (v < 0 || v >= MAX_VMON) {
		flog(LOG_WARN, "(write) vmon inesistente: %d", v);
		abort_p();
	}

	des_vmon* t = &array_desvmon[v];
	int max_off = t->ch_x * t->ch_y;
	if (quanti <= 0 || off >= max_off || off + quanti <= 0)
		return;

	if (off < 0) {
		vetti += -off;
		quanti -= -off;
		off = 0;
	}
	if (off + quanti >= max_off) 
		quanti = max_off - off;
	for (int i = 0; i < quanti; i++)
		vmon_put(v, off + i, vetti[i]);
}
void light_to(int v)
{
	des_vmon* t = &array_desvmon[v];

	if (vmon_active) 
		color_bars(vmon_active, 0x09);
	color_bars(t, 0x0B);

}

extern "C" void c_vmon_switch(int v)
{
	if (v < 0 || v >= MAX_VMON) {
		flog(LOG_WARN, "(switch) vmon inesistente: %d", v);
		abort_p();
	}
	if (framebuffer) {
		light_to(v);
		vmon_active = &array_desvmon[v];
	} else {
		if (vmon_active) {
			memcpy(vmon_active->buf, VIDEO_MEM_BASE, VIDEO_MEM_SIZE * 2);
			vmon_active->video = vmon_active->buf;
		}
		vmon_active = &array_desvmon[v];
		vmon_active->video = VIDEO_MEM_BASE;
		memcpy(VIDEO_MEM_BASE, vmon_active->buf, VIDEO_MEM_SIZE * 2);
		console_set_cursor_size(vmon_active->cursor_size);
		console_cursor(vmon_active->cursor);
	}
}

extern "C" bool c_vmon_getsize(int v, natw& maxx, natw& maxy)
{
	if (v < 0 || v >= MAX_VMON) {
		return false;
	}
	des_vmon *t = &array_desvmon[v];
	maxx = t->ch_x;
	maxy = t->ch_y;
	return true;
}

			
	
extern "C" void c_vmon_setcursor(int v, natl off)
{
	if (v < 0 || v >= MAX_VMON) {
		flog(LOG_WARN, "(cursor) vmon inesistente: %d", v);
		abort_p();
	}
	des_vmon *t = &array_desvmon[v];
	if (t == vmon_active) {
		if (framebuffer) {
			framebuffer_update_cursor(t, off);
		} else {
			t->cursor = off;
			console_cursor(t->cursor);
		}
	}
}


extern "C" void c_vmon_cursor_shape(int v, int shape)
{
	if (v < 0 || v >= MAX_VMON) {
		flog(LOG_WARN, "(cshape) vmon inesistente: %d", v);
		abort_p();
	}

	int size = 14;
	if (shape == 1) 
		size = 2;
	des_vmon *t = &array_desvmon[v];
	if (framebuffer)
		framebuffer_cursor(t);
	t->cursor_size = size;
	if (framebuffer)
		framebuffer_cursor(t);
	else
		console_set_cursor_size(t->cursor_size);

}

////////////////////////////////////////////////////////////////////////////////
//                         GESTIONE DELLA CONSOLE [9.5]                       //
////////////////////////////////////////////////////////////////////////////////

natw COLS = 80; 	// [9.5]
natw ROWS = 25;	// [9.5]
natl VIDEO_SIZE = COLS * ROWS;	// [9.5]

struct des_vid {
	int vmon;
	natl x, y;
	natb attr;
};

const natl MAX_CODE = 29; // [9.5]

struct des_ckbd { // [9.5]
	int vkbd;
	addr punt;
	natl cont;
	bool shift;
	natb tab[MAX_CODE];
	natb tabmin[MAX_CODE];
	natb tabmai[MAX_CODE];
};

struct des_console { // [9.5]
	natl mutex;
	natl sincr;
	des_ckbd kbd;
	des_vid vid;
};

extern "C" des_console console; // [9.5]

void scroll(des_vid *p_des)	// [9.5]
{
	int vmon = p_des->vmon;
	for (int i = 0; i < VIDEO_SIZE - COLS; i++) 
		vmon_put(vmon, i, vmon_get(vmon, i + COLS));
	for (natl i = 0; i < COLS; i++)
		vmon_put(vmon, VIDEO_SIZE - COLS + i, 0 | p_des->attr << 8);
	p_des->y--;
}

void writeelem(natb c) {	// [9.5]
	des_vid* p_des = &console.vid;
	int vmon = p_des->vmon;
	switch (c) {
	case '\r':
		p_des->x = 0;
		break;
	case '\n':
		p_des->y++;
		if (p_des->y >= ROWS)
			scroll(p_des);
		break;
	case '\b':
		if (p_des->x > 0 || p_des->y > 0) {
			if (p_des->x == 0) {
				p_des->x = COLS - 1;
				p_des->y--;
			} else
				p_des->x--;
		}
		break;
	default:
		vmon_put(vmon, p_des->y * COLS + p_des->x, c | p_des->attr << 8);
		p_des->x++;
		if (p_des->x >= COLS) {
			p_des->x = 0;
			p_des->y++;
		}
		if (p_des->y >= ROWS) 
			scroll(p_des);
		break;
	}
	c_vmon_setcursor(vmon, p_des->y * COLS + p_des->x);
}

void writeseq(cstr seq)	// [9.5]
{
	const natb* pn = static_cast<const natb*>(seq);
	while (*pn != 0) {
		writeelem(*pn);
		pn++;
	}
}

extern "C" void c_writeconsole(cstr buff) // [9.5]
{
	des_console *p_des = &console;
	sem_wait(p_des->mutex);
	writeseq(buff);
	writeseq("\r\n");
	sem_signal(p_des->mutex);
}

//extern "C" void go_inputkbd(interfkbd_reg indreg); // [9.5]
//extern "C" void halt_inputkbd(interfkbd_reg indreg); // [9.5]

void startkbd_in(des_ckbd* p_des, str buff) // [9.5]
{
	p_des->punt = buff;
	p_des->cont = 80;
	//go_inputkbd(p_des->indreg);
	c_vkbd_intr_enable(p_des->vkbd, true);
}

extern "C" void c_readconsole(str buff, natl& quanti) // [9.5]
{
	des_console *p_des;

	p_des = &console;
	sem_wait(p_des->mutex);
	startkbd_in(&p_des->kbd, buff);
	sem_wait(p_des->sincr);
	quanti = p_des->kbd.cont;
	sem_signal(p_des->mutex);
}

natb converti(des_ckbd* p_des, natb c) { // [9.5]
	natb cc;
	natl pos = 0;
	while (pos < MAX_CODE && p_des->tab[pos] != c)
		pos++;
	if (pos == MAX_CODE)
		return 0;
	if (p_des->shift)
		cc = p_des->tabmai[pos];
	else
		cc = p_des->tabmin[pos];
	return cc;
}

void estern_ckbd(int h) // [9.5]
{
	des_console *p_des = &console;
	natw C;
	natb a, c;
	bool fine;

	for(;;) {
		//halt_inputkbd(p_des->kbd.indreg);
		c_vkbd_intr_enable(p_des->kbd.vkbd, false);

		//inputb(p_des->kbd.indreg.iRBR, c);
		C = c_vkbd_read(p_des->kbd.vkbd);
		if (C == 0xFFFF)
			terminate_p();

		c = (natb)C;
		
		fine = false;
		switch (c) {
		case 0x2a: // left shift make code
			p_des->kbd.shift = true;
			break;
		case 0xaa: // left shift break code
			p_des->kbd.shift = false;
			break;
		default:
			if (c < 0x80) {
				a = converti(&p_des->kbd, c);
				if (a == 0)
					break;
				if (a == '\b') {
					if (p_des->kbd.cont < 80) {
						p_des->kbd.punt = static_cast<natb*>(p_des->kbd.punt) - 1;
						p_des->kbd.cont++;
						writeseq("\b \b");
					}
				} else if (a == '\r' || a == '\n') {
					fine = true;
					p_des->kbd.cont = 80 - p_des->kbd.cont;
					*static_cast<natb*>(p_des->kbd.punt) = 0;
					writeseq("\r\n");
				} else {
					*static_cast<natb*>(p_des->kbd.punt) = a;
					p_des->kbd.punt = static_cast<natb*>(p_des->kbd.punt) + 1;
					p_des->kbd.cont--;
					writeelem(a);
					if (p_des->kbd.cont == 0) {
						fine = true;
						p_des->kbd.cont = 80;
					}
				}
			}
			break;
		}
		if (fine == true) 
			sem_signal(p_des->sincr);
		else
			//go_inputkbd(p_des->kbd.indreg);
			c_vkbd_intr_enable(p_des->kbd.vkbd, true);
		//nwfi(master);
		c_vkbd_wfi(p_des->kbd.vkbd);
	}
}

// (* inizializzazioni
bool con_vid_init(int vmon);

extern "C" void c_iniconsole(natb cc)
{
	des_vid *p_des = &console.vid;
	p_des->attr = cc;
	int vmon = p_des->vmon;
	for (int i = 0; i < VIDEO_SIZE; i++) 
		vmon_put(vmon, i, i | p_des->attr << 8);
	p_des->x = p_des->y = 0;
	c_vmon_setcursor(vmon, p_des->y * COLS + p_des->x);
}

bool con_kbd_init(int vkbd)
{
	console.kbd.vkbd = vkbd;
	if (activate_p(estern_ckbd, 0, PRIO, LIV) == 0xFFFFFFFF) {
		flog(LOG_ERR, "console: impossibile creare estern_ckbd");
		return false;
	}
	return true;
}

extern "C" des_vid vid;

bool con_vid_init(int vmon)
{
	c_vmon_getsize(vmon, COLS, ROWS);
	VIDEO_SIZE = COLS * ROWS;
	des_vid *p_des = &console.vid;
	p_des->vmon = vmon;
	for (int i = 0; i < VIDEO_SIZE; i++) 
		vmon_put(vmon, i, 0 | p_des->attr << 8);
	c_vmon_setcursor(vmon, p_des->y * COLS + p_des->x);
	flog(LOG_INFO, "vid: video inizializzato");
	return true;
}

bool console_init(int vkbd, int vmon) {
	des_console *p_des = &console;

	if ( (p_des->mutex = sem_ini(1)) == 0xFFFFFFFF) {
		flog(LOG_ERR, "console: impossibile creare mutex");
		return false;
	}
	if ( (p_des->sincr = sem_ini(0)) == 0xFFFFFFFF) {
		flog(LOG_ERR, "console: impossibile creare sincr");
		return false;
	}
	return con_kbd_init(vkbd) && con_vid_init(vmon);
}

// *)

/////////////////////////////////////////////////////////////////////////////////
//                  FUNZIONI DI LIBRERIA                                       //
/////////////////////////////////////////////////////////////////////////////////
typedef char *va_list;

// Versione semplificata delle macro per manipolare le liste di parametri
//  di lunghezza variabile; funziona solo se gli argomenti sono di
//  dimensione multipla di 4, ma e' sufficiente per le esigenze di printk.
//
#define va_start(ap, last_req) (ap = (char *)&(last_req) + sizeof(last_req))
#define va_arg(ap, type) ((ap) += sizeof(type), *(type *)((ap) - sizeof(type)))
#define va_end(ap)

natl strlen(const char *s)
{
	natl l = 0;

	while(*s++)
		++l;

	return l;
}

char *strncpy(char *dest, const char *src, unsigned long l)
{
	unsigned long i;

	for(i = 0; i < l && src[i]; ++i)
		dest[i] = src[i];

	return dest;
}

static const char hex_map[16] = { '0', '1', '2', '3', '4', '5', '6', '7',
	'8', '9', 'a', 'b', 'c', 'd', 'e', 'f' };

static void htostr(char *buf, unsigned long l)
{
	int i;

	buf[0] = '0';
	buf[1] = 'x';

	for(i = 9; i > 1; --i) {
		buf[i] = hex_map[l % 16];
		l /= 16;
	}
}

static void itostr(char *buf, unsigned int len, long l)
{
	natl i, div = 1000000000, v, w = 0;

	if(l == (-2147483647 - 1)) {
		strncpy(buf, "-2147483648", 12);
		return;
	} else if(l < 0) {
		buf[0] = '-';
		l = -l;
		i = 1;
	} else if(l == 0) {
		buf[0] = '0';
		buf[1] = 0;
		return;
	} else
		i = 0;

	while(i < len - 1 && div != 0) {
		if((v = l / div) || w) {
			buf[i++] = '0' + (char)v;
			w = 1;
		}

		l %= div;
		div /= 10;
	}

	buf[i] = 0;
}

#define DEC_BUFSIZE 12

int vsnprintf(char *str, natl size, const char *fmt, va_list ap)
{
	natl in = 0, out = 0, tmp;
	char *aux, buf[DEC_BUFSIZE];

	while(out < size - 1 && fmt[in]) {
		switch(fmt[in]) {
			case '%':
				switch(fmt[++in]) {
					case 'd':
						tmp = va_arg(ap, int);
						itostr(buf, DEC_BUFSIZE, tmp);
						if(strlen(buf) >
								size - out - 1)
							goto end;
						for(aux = buf; *aux; ++aux)
							str[out++] = *aux;
						break;
					case 'x':
						tmp = va_arg(ap, int);
						if(out > size - 11)
							goto end;
						htostr(&str[out], tmp);
						out += 10;
						break;
					case 's':
						aux = va_arg(ap, char *);
						while(out < size - 1 && *aux)
							str[out++] = *aux++;
						break;	
				}
				++in;
				break;
			default:
				str[out++] = fmt[in++];
		}
	}
end:
	str[out++] = 0;

	return out;
}

int snprintf(char *buf, unsigned long n, const char *fmt, ...)
{
	va_list ap;
	int l;

	va_start(ap, fmt);
	l = vsnprintf(buf, n, fmt, ap);
	va_end(ap);

	return l;
}

// copia n byte da src a dest
void *memcpy(void *dest, const void *src, unsigned int n)
{
	char       *dest_ptr = static_cast<char*>(dest);
	const char *src_ptr  = static_cast<const char*>(src);

	if (src_ptr < dest_ptr && src_ptr + n > dest_ptr)
		for (int i = n - 1; i >= 0; i--)
			dest_ptr[i] = src_ptr[i];
	else
		for (natl i = 0; i < n; i++)
			dest_ptr[i] = src_ptr[i];

	return dest;
}

// scrive n byte pari a c, a partire da dest
void *memset(void *dest, int c, unsigned int n)
{
	char *dest_ptr = static_cast<char*>(dest);

        for (natl i = 0; i < n; i++)
              dest_ptr[i] = static_cast<char>(c);

        return dest;
}

// log formattato
void flog(log_sev sev, const char *fmt, ...)
{
	va_list ap;
	char buf[LOG_MSG_SIZE];

	va_start(ap, fmt);
	int l = vsnprintf(buf, LOG_MSG_SIZE, fmt, ap);
	va_end(ap);

	if (l > 1)
		log(sev, buf, l - 1);
}

////////////////////////////////////////////////////////////////////////////////
//                 INIZIALIZZAZIONE DEL SOTTOSISTEMA DI I/O                   //
////////////////////////////////////////////////////////////////////////////////

// inizializza i gate usati per le chiamate di IO
//
extern "C" void fill_io_gates(void);

// eseguita in fase di inizializzazione
//
extern "C" void cmain(int sem_io)
{
	fill_io_gates();
	
	if (!eth_init()) {
		flog(LOG_WARN, "eth_init fallita");
		abort_p();
	}
	
	if (!kbd_init()) {
		flog(LOG_WARN, "kdb_init fallita");
		abort_p();
	}
	if (!vkbd_init()) {
		flog(LOG_WARN, "vkdb_init fallita");
		abort_p();
	}
	if (!com_init()) {
		flog(LOG_WARN, "com_init fallita");
		abort_p();
	}
	if (!vmon_init()) {
		flog(LOG_WARN, "vmon_init fallita");
		abort_p();
	}
	if (!console_init(VKBD_CONSOLE, 0)) {
		flog(LOG_WARN, "console_init fallita");
		abort_p();
	}
	sem_signal(sem_io);
	terminate_p();
}
